% If output D is specified, it will store the history of iterations for each
% point (cell array of m x 2 arrays, one array per point)

function [u, v, D]= lucas_kanade(I1, I2, pos, winsize, prediction)

% Does the user want us to keep track of iteration history?
if nargout > 2
    history = 1;
    D = {};
else
    history = 0;
end

% Size of square window can be specified. Must be odd.
if nargin < 4
    % Size of tracking window
    winsize = 15 * [1 1];
else
    if mod(winsize, 2) == 0
        error('Window size must be odd')
    end
    winsize = winsize * [1 1];
end

if nargin < 5
    prediction = zeros(size(pos));
end

% Maximum number of iterations
maxIter = 100;

% Desired precision in the result (pixels)
precision = 0.1;

% Number of points
k = size(pos,1);

% Half the window size
half = floor(winsize(1)/2);

% Motions greater than this are suspicious
maxMotion = winsize;

[Ir, Ic] = gradient(I1);

% Square root of a Gaussian filter
g = gauss(-half:half, 0, half/2, 1);
w = g(:) * g(:)';
w = sqrt(w(:));

for i = 1:k
    % Skip lost features
    if isnan(pos(i, 1))
        u(i) = NaN;
        v(i) = NaN;
        continue;
    end
 
    rows0 = pos(i,1)+[-half:half];
    cols0 = pos(i,2)+[-half:half];
    
    % Stay inside the image (assumes that motion is less than 'half')
    if min(rows0) < half | min(cols0) < half | ...
            max(rows0) > size(I1, 1) - half | max(cols0) > size(I1, 2) - half
        u(i) = 0;
        v(i) = 0;
    else
        % Derivatives in the first window are computed only once
        Iwin = interp2(I1, cols0, rows0');
        r = interp2(Ir, cols0, rows0');
        c = interp2(Ic, cols0, rows0');
        
        % Ditto for the left-hand side of the normal equations
        G = [w.*r(:) w.*c(:)];
        A = G;
       
        d = prediction(i, :)';
        
        if history
            D{i} = d';
        end
        
        for m = 1:maxIter
                      
            % Subpixel interpolation for the second frame
            rows = rows0 + d(1);
            cols = cols0 + d(2);
            t = interp2(I2, cols, rows') - Iwin;
          
            % Right-hand side of the normal equations
            b = - w .* t(:);
            
            % Solve A*x = b
            x = A\b;
            
            d = d + x;
            
            if history
                D{i} = [D{i}; d'];
            end
            
            if norm(x) < precision
                break;
            end
            
            if norm(d) > maxMotion
                % Probably lost
                d = [NaN; NaN];
                break;
            end
        end
        
        if m >= maxIter
            disp(sprintf('Warning: %d iterations insufficient for %f pixel precision', ...
                m, precision))
        end
        
        u(i)=d(1);
        v(i)=d(2);
    end
end